from collections import OrderedDict, abc
from enum import Enum
from functools import reduce, wraps
from typing import Callable, List, Optional, Sequence, Tuple, TypeVar, Union, overload

import torch
from torch import Tensor, nn

from .delay import Delay
from .logging import getLogger
from .module import CoModule, PaddingMode, _callmode
from .skip import Skip
from .utils import (
    function_repr,
    load_state_dict,
    num_from,
    state_dict,
    temporary_parameter,
)

logger = getLogger(__name__)

__all__ = [
    "Sequential",
    "BroadcastReduce",
    "Residual",
    "Broadcast",
    "Parallel",
    "ParallelDispatch",
    "Reduce",
    "Conditional",
]

T = TypeVar("T")
S = TypeVar("S")

State = List[Optional[Tensor]]


class Reduction(Enum):
    """Types of parallel tensor reduce operation.
    Supported tupes are:
    - SUM:    Element-wise summation
    - CONCAT: Channel-wise concatenation
    - MUL:    Hadamark product
    """

    SUM = "sum"
    CONCAT = "concat"
    MUL = "mul"
    MAX = "max"


ReductionFunc = Callable[[Sequence[Tensor]], Tensor]
ReductionFuncOrEnum = Union[Reduction, ReductionFunc, str]


def apply_forward(module: CoModule, input: Tensor):
    if isinstance(module, nn.RNNBase):
        return module.forward(input)[0]
    return module(input)


def reduce_sum(inputs: Sequence[Tensor]) -> Tensor:
    assert len(inputs) >= 2
    return reduce(torch.Tensor.add, inputs[1:], inputs[0])


def reduce_concat(inputs: Sequence[Tensor]) -> Tensor:
    """Channel-wise concatenation of input

    Args:
        inputs (Sequence[Tensor]): Inputs with broadcastable shapes.

    Returns:
        Tensor: Inputs concatenated in the channel dimension
    """
    return torch.cat(inputs, dim=1)  # channel dim for inputs of shape (B, C, T, H, W)


def reduce_mul(inputs: Sequence[Tensor]) -> Tensor:
    """Hadamard product between inputs

    Args:
        inputs (Sequence[Tensor]): Inputs with broadcastable shapes.

    Returns:
        Tensor: Haramard product of inputs
    """
    assert len(inputs) >= 2
    return reduce(torch.Tensor.mul, inputs[1:], inputs[0])


def reduce_max(inputs: Sequence[Tensor]) -> Tensor:
    assert len(inputs) >= 2
    return reduce(torch.max, inputs[1:], inputs[0])


def nonempty(fn: ReductionFunc) -> ReductionFunc:
    @wraps(fn)
    def wrapped(inputs: Sequence[Tensor]) -> Tensor:
        if any(inp is None or inp.shape[0] == 0 for inp in inputs):
            return None  # pragma: no cover
        return fn(inputs)

    return wrapped


class FlattenableStateDict:
    """Mixes in the ability to flatten state dicts.
    It is assumed that classes that inherit this module also inherit from nn.Module
    """

    flatten_state_dict = False

    def __init__(self, *args, **kwargs):
        ...  # pragma: no cover

    def state_dict(
        self, destination=None, prefix="", keep_vars=False, flatten=False
    ) -> "OrderedDict[str, Tensor]":
        flatten = flatten or self.flatten_state_dict
        return state_dict(self, destination, prefix, keep_vars, flatten)

    def load_state_dict(
        self,
        state_dict: "OrderedDict[str, Tensor]",
        strict: bool = True,
        flatten=False,
    ):
        flatten = flatten or self.flatten_state_dict
        return load_state_dict(self, state_dict, strict, flatten)


def co_add_module(self, name: str, module: Optional["nn.Module"]) -> None:
    if not CoModule.is_valid(module):
        # Attempt automatic conversion
        from continual.convert import continual  # break cyclical import

        module = continual(module)

    nn.Module.add_module(self, name, module)


class Broadcast(CoModule, nn.Module):
    """Broadcast one input stream to multiple output streams.

    This is needed for handling parallel streams in subsequent modules.
    For instance, here is how it is used to create a residual connection::

        residual = co.Sequential(
            co.Broadcast(2),
            co.Parallel(
                co.Conv3d(32, 32, kernel_size=3, padding=1),
                co.Delay(2),
            ),
            co.Reduce("sum"),
        )

    Since the ``Broadcast`` -> ``Parallel`` -> ``Reduce`` sequence is so common,
    identical behavior can be achieved with ``BroadcastReduce`` ::

        residual = co.BroadcastReduce(
            co.Conv3d(32, 32, kernel_size=3, padding=1),
            co.Delay(2),
            reduce="sum"
        )

    Even shorter, the library features a residual connection, which automatically handles delays::

        residual = co.Residual(co.Conv3d(32, 32, kernel_size=3, padding=1))

    Args:
        num_streams (int):
            Number of streams to broadcast to. If none are given, a Sequential
            module may infer it automatically.
    """

    _state_shape = 0
    _dynamic_state_inds = []

    def __init__(
        self,
        num_streams: int = None,
    ):
        nn.Module.__init__(self)
        self.num_streams = num_streams

    def forward(self, input: T) -> List[T]:
        assert isinstance(
            self.num_streams, int
        ), "Unknown number of target streams in Broadcast."
        return [input for _ in range(self.num_streams)]

    def _forward_step(self, input: T, prev_state=None):
        return self.forward(input), prev_state

    def forward_step(self, input: T, update_state=True) -> List[T]:
        return self.forward(input)

    def forward_steps(self, input: T, pad_end=False, update_state=True) -> List[T]:
        return self.forward(input)


class Parallel(FlattenableStateDict, CoModule, nn.Sequential):
    """Container for  executing modules in parallel.
    Modules will be added to it in the order they are passed in the
    constructor.

    For instance, here is how it is used to create a residual connection::

        residual = co.Sequential(
            co.Broadcast(2),
            co.Parallel(
                co.Conv3d(32, 32, kernel_size=3, padding=1),
                co.Delay(2),
            ),
            co.Reduce("sum"),
        )

    Since the ``Broadcast`` -> ``Parallel`` -> ``Reduce`` sequence is so common,
    identical behavior can be achieved with ``BroadcastReduce`` ::

        residual = co.BroadcastReduce(
            co.Conv3d(32, 32, kernel_size=3, padding=1),
            co.Delay(2),
            reduce="sum"
        )

    Even shorter, the library features a residual connection, which automatically handles delays::

        residual = co.Residual(co.Conv3d(32, 32, kernel_size=3, padding=1))


    Args:
        arg (OrderedDict[str, CoModule]): An OrderedDict of strings and modules.
        *args (CoModule): Comma-separated modules.
        auto_delay (bool, optional):
            Automatically add delay to modules in order to match the longest delay.
            Defaults to True.
    """

    @overload
    def __init__(
        self,
        *args: CoModule,
        auto_delay=True,
    ) -> None:
        ...  # pragma: no cover

    @overload
    def __init__(
        self,
        arg: "OrderedDict[str, CoModule]",
        auto_delay=True,
    ) -> None:
        ...  # pragma: no cover

    def __init__(
        self,
        *args,
        auto_delay=True,
    ):
        nn.Module.__init__(self)

        if len(args) == 1 and isinstance(args[0], OrderedDict):
            modules = [(key, module) for key, module in args[0].items()]
        else:
            modules = [(str(idx), module) for idx, module in enumerate(args)]

        if auto_delay:
            # If there is a delay mismatch, automatically add delay to match the longest
            max_delay = max([m.delay for _, m in modules])
            modules = [
                (
                    key,
                    (
                        Sequential(module, Delay(max_delay - module.delay))
                        if module.delay < max_delay
                        else module
                    ),
                )
                for key, module in modules
            ]

        assert (
            len(set(num_from(getattr(m, "stride", 1)) for _, m in modules)) == 1
        ), f"Expected all modules to have the same stride, but got strides {[(num_from(getattr(m, 'stride', 1))) for _, m in modules]}"

        for key, module in modules:
            self.add_module(key, module)

        delays = set(m.delay for m in self)
        if len(delays) != 1:  # pragma: no cover
            logger.warning(
                f"It recommended that parallel modules have the same delay, but found delays {delays}. "
                "Temporal consistency cannot be guaranteed."
            )
        self._delay = max(delays)

        receptive_fields = set(m.receptive_field for m in self)
        self._receptive_field = max(receptive_fields)

    def add_module(self, name: str, module: Optional["nn.Module"]) -> None:
        co_add_module(self, name, module)

    @property
    def _state_shape(self):
        return [m._state_shape for m in self]

    @property
    def _dynamic_state_inds(self):
        return [m._dynamic_state_inds for m in self]

    def _forward_step(
        self, inputs: List[T], prev_state: Optional[List[Optional[S]]] = None
    ):
        prev_state = prev_state or [None for _ in range(len(self))]
        outs, next_state = [], []
        for i, module in enumerate(self):
            out, n_state = module._forward_step(inputs[i], prev_state=prev_state[i])
            outs.append(out)
            next_state.append(n_state)
        return outs, next_state

    def forward_step(self, inputs: List[T], update_state=True) -> List[T]:
        outs = []
        for i, m in enumerate(self):
            with temporary_parameter(m, "call_mode", _callmode("forward_step")):
                outs.append(m(inputs[i], update_state=update_state))
        return outs

    def forward_steps(
        self, inputs: List[T], pad_end=False, update_state=True
    ) -> List[T]:
        outs = []
        for i, m in enumerate(self):
            with temporary_parameter(m, "call_mode", _callmode("forward_steps")):
                outs.append(m(inputs[i], pad_end=pad_end, update_state=update_state))
        return outs

    def forward(self, inputs: List[T]) -> List[T]:
        outs = []
        for i, m in enumerate(self):
            with temporary_parameter(m, "call_mode", _callmode("forward")):
                outs.append(apply_forward(m, inputs[i]))
        return outs

    @property
    def receptive_field(self) -> int:
        return self._receptive_field

    @property
    def delay(self) -> int:
        return self._delay

    @property
    def stride(self) -> Tuple[int]:
        return getattr(next(iter(self)), "stride", (1,))

    @property
    def padding(self) -> Tuple[int]:
        return (max(getattr(m, "padding", (0,))[0] for m in self),)

    def clean_state(self):
        for m in self:
            if hasattr(m, "clean_state"):
                m.clean_state()


class ParallelDispatch(CoModule, nn.Module):
    """Reorder, copy, and group streams from parallel streams.

    Reorder example::

        net = co.Sequential(
            co.Broadcast(2),
            co.Parallel(co.Add(1), co.Identity()),
            co.ParallelDispatch([1,0]),  # Reorder stream 0 and 1
            co.Parallel(co.Identity(), co.Add(2)),
            co.Reduce("max"),
        )

        assert torch.equal(net(torch.tensor([0])), torch.tensor([3]))

    Depiction of the reorder example::

               | -> co.Add(1)     \\ / -> co.Identity() |
        [0] -> |                   X                   | -> max -> [3]
               | -> co.Identity() /  \\ -> co.Add(2)    |

    Copy example::

        net = co.Sequential(
            co.Broadcast(2),
            co.Parallel(co.Add(1), co.Identity()),
            co.ParallelDispatch([0, 0, 1]),  # Copy stream 0
            co.Parallel(co.Identity(), co.Add(2), co.Identity()),
            co.Reduce("max"),
        )

        assert torch.equal(net(torch.tensor([0])), torch.tensor([3]))

    Depiction of the copy example::

               | -> co.Add(1)  -> | -> co.Identity() -> |
        [0] -> |                  | -> co.Add(2)     -> | -> max -> [3]
               | -> co.Identity() ------> co.Add(1)  -> |

    Group example::

        net = co.Sequential(
            co.Broadcast(2),
            co.Parallel(co.Add(2), co.Identity()),
            co.ParallelDispatch([[0, 0], 1]),  # Copy and group stream 0
            co.Parallel(co.Reduce("sum"), co.Identity()),
            co.Reduce("max"),
        )

        assert torch.equal(net(torch.tensor([0])), torch.tensor([4]))

    Depiction of the group example::

                                 | -> |
               | -> co.Add(2) -> |    | ->    sum     -> |
        [0] -> |                 | -> |                  | -> max -> [4]
               | -> co.Identity() ----> co.Identity() -> |

    Args:
        dispatch_mapping (Sequence[Union[int, Sequence[int]]]):
            input-to-output mapping, where the integers signify the input stream ordering
            and the positions denote corresponding output ordering.
            Examples::
                [1,0] to shuffle order of streams.
                [0,1,1] to copy stream 1 onto a new stream.
                [[0,1],2] to group stream 0 and 1 while keeping stream 2 separate.

    """

    _state_shape = 0
    _dynamic_state_inds = []

    def __init__(
        self,
        dispatch_mapping: Sequence[Union[int, Sequence[int]]],
    ):
        nn.Module.__init__(self)

        def is_int_or_valid_list(x):
            if isinstance(x, int):
                return True
            elif isinstance(x, abc.Sequence):
                return all(is_int_or_valid_list(z) for z in x)
            else:
                return False

        assert isinstance(dispatch_mapping, abc.Sequence) and is_int_or_valid_list(
            dispatch_mapping
        ), "The dispatch_mapping should be of type Sequence[Union[StreamId, Sequence[StreamId]]]"

        self.dispatch_mapping = dispatch_mapping

    def forward(self, input: List[T]) -> List[Union[T, List[T]]]:
        def dispatch(mapping):
            nonlocal input
            if isinstance(mapping, abc.Sequence):
                return [dispatch(m) for m in mapping]
            else:
                return input[mapping]

        return dispatch(self.dispatch_mapping)

    def _forward_step(self, input: List[T], prev_state=None):
        return self.forward_step(input), prev_state

    def forward_step(
        self, input: List[T], update_state=True
    ) -> List[Union[T, List[T]]]:
        return self.forward(input)

    def forward_steps(
        self, input: List[T], pad_end=False, update_state=True
    ) -> List[Union[T, List[T]]]:
        return self.forward(input)


class Reduce(CoModule, nn.Module):
    """Reduce multiple input streams to a single using the selected function

    For instance, here is how it is used to sum streams in a residual connection::

        residual = co.Sequential(
            co.Broadcast(2),
            co.Parallel(
                co.Conv3d(32, 32, kernel_size=3, padding=1),
                co.Delay(2),
            ),
            co.Reduce("sum"),
        )

    A user-defined can be passed as well::

        from functools import reduce

        def my_sum(inputs):
            return reduce(torch.Tensor.add, inputs[1:], inputs[0])

        residual = co.Sequential(
            co.Broadcast(2),
            co.Parallel(
                co.Conv3d(32, 32, kernel_size=3, padding=1),
                co.Delay(2),
            ),
            co.Reduce(my_sum),
        )

    Args:
        reduce (Union[str, Callable[[Sequence[Tensor]], Tensor]]):
            Reduce function. Either one of ["sum", "channel", "mul", "max"] or
            user-defined function mapping a sequence of tensors to a single one.

    """

    _state_shape = 0
    _dynamic_state_inds = []

    def __init__(
        self,
        reduce: ReductionFuncOrEnum = "sum",
    ):
        nn.Module.__init__(self)
        self.reduce = nonempty(
            reduce
            if callable(reduce)
            else {
                Reduction.SUM: reduce_sum,
                Reduction.CONCAT: reduce_concat,
                Reduction.MUL: reduce_mul,
                Reduction.MAX: reduce_max,
            }[Reduction(reduce)]
        )

    def forward(self, inputs: List[T]) -> T:
        return self.reduce(inputs)

    def _forward_step(self, inputs: List[T], prev_state=None):
        if all(isinstance(i, Tensor) for i in inputs):
            return self.reduce(inputs), prev_state
        return None, prev_state  # pragma: no cover

    def forward_step(self, inputs: List[T], update_state=True) -> T:
        return self._forward_step(inputs)[0]

    def forward_steps(self, inputs: List[T], pad_end=False, update_state=True) -> T:
        return self.reduce(inputs)


class Sequential(FlattenableStateDict, CoModule, nn.Sequential):
    """A sequential container.
    This module is an augmentation of `torch.nn.Sequential`
    which adds continual inference methods

    Modules will be added to it in the order they are passed in the
    constructor. Alternatively, an ``OrderedDict`` of modules can be
    passed in. The ``forward()``, ``forward_step()`` and ``forward_steps()``
    methods of ``Sequential`` accept any input and forwards it to the first
    module it contains. It then "chains" outputs to inputs sequentially for
    each subsequent module, finally returning the output of the last module.

    The value a ``Sequential`` provides over manually calling a sequence
    of modules is that it allows treating the whole container as a
    single module, such that performing a transformation on the
    ``Sequential`` applies to each of the modules it stores (which are
    each a registered submodule of the ``Sequential``).

    Example::

        # Using Sequential to create a small model. When `model` is run,
        # input will first be passed to `Conv2d(1,20,5)`. The output of
        # `Conv2d(1,20,5)` will be used as the input to the first
        # `ReLU`; the output of the first `ReLU` will become the input
        # for `Conv2d(20,64,5)`. Finally, the output of
        # `Conv2d(20,64,5)` will be used as input to the second `ReLU`
        model = co.Sequential(
            co.Conv2d(1,20,5),
            nn.ReLU(),
            co.Conv2d(20,64,5),
            nn.ReLU()
        )

        # Using Sequential with OrderedDict. This is functionally the
        # same as the above code
        model = co.Sequential(OrderedDict([
            ('conv1', co.Conv2d(1,20,5)),
            ('relu1', nn.ReLU()),
            ('conv2', co.Conv2d(20,64,5)),
            ('relu2', nn.ReLU())
        ]))
    """

    @overload
    def __init__(self, *args: nn.Module) -> None:
        ...  # pragma: no cover

    @overload
    def __init__(self, arg: "OrderedDict[str, nn.Module]") -> None:
        ...  # pragma: no cover

    def __init__(self, *args):
        nn.Module.__init__(self)
        modules = []
        if len(args) == 1 and isinstance(args[0], OrderedDict):
            for key, module in args[0].items():
                modules.append((key, module))
        else:
            for idx, module in enumerate(args):
                modules.append((str(idx), module))

        # If a co.Broadcast is followed by a co.Parallel, automatically infer num_streams
        for i in range(len(modules)):
            if isinstance(modules[i][1], Broadcast) and i < len(modules) - 1:
                if isinstance(modules[i + 1][1], Parallel):
                    modules[i][1].num_streams = modules[i][1].num_streams or len(
                        modules[i + 1][1]
                    )

        for n, m in modules:
            self.add_module(n, m)

    def add_module(self, name: str, module: Optional["nn.Module"]) -> None:
        co_add_module(self, name, module)

    def forward(self, input):
        for m in self:
            with temporary_parameter(m, "call_mode", _callmode("forward")):
                input = apply_forward(m, input)
        return input

    def forward_step(self, input, update_state=True):
        for module in self:
            # ptflops only works when __call__ is triggered
            with temporary_parameter(module, "call_mode", _callmode("forward_step")):
                input = module(
                    input, update_state=update_state
                )  # == module.forward_step
            if not type(input) in {Tensor, list}:
                return None
        return input

    def _forward_step(self, input: torch.Tensor, prev_state: List[State]):
        prev_state = prev_state or [None for _ in range(len(self))]
        next_state = prev_state.copy()
        for i, module in enumerate(self):
            input, n_state = module._forward_step(input, prev_state=prev_state[i])
            next_state[i] = n_state
            if input is None:
                return None, next_state
        return input, next_state

    @property
    def _state_shape(self):
        return [m._state_shape for m in self]

    @property
    def _dynamic_state_inds(self):
        return [m._dynamic_state_inds for m in self]

    def forward_steps(self, input: Tensor, pad_end=False, update_state=True):
        for m in self:
            if not type(input) in {Tensor, list} or len(input) == 0:
                return None  # pragma: no cover
            # ptflops only works when __call__ is triggered
            with temporary_parameter(m, "call_mode", _callmode("forward_steps")):
                # == m.forward_steps
                input = m(input, pad_end=pad_end, update_state=update_state)

        return input

    @property
    def receptive_field(self) -> int:
        reverse_modules = [m for m in self][::-1]
        rf = reverse_modules[0].receptive_field
        for m in reverse_modules[1:]:
            s = getattr(m, "stride", [1])
            s = s[0]
            rf = s * rf + m.receptive_field - s
        return rf

    @property
    def stride(self) -> Tuple[int]:
        tot = 1
        for m in self:
            tot *= m.stride[0]
        return (tot,)

    @property
    def padding(self) -> Tuple[int]:
        m = [m for m in self]
        p = m[0].padding[0]
        s = m[0].stride[0]
        for i in range(1, len(m)):
            p += m[i].padding[0] * s
            s = s * m[i].stride[0]
        return (p,)

    @staticmethod
    def build_from(module: nn.Sequential) -> "Sequential":
        from .convert import continual  # import here due to circular import

        return Sequential(
            OrderedDict([(k, continual(m)) for k, m in module._modules.items()])
        )

    def clean_state(self):
        for m in self:
            if hasattr(m, "clean_state"):
                m.clean_state()

    def append(self, module: nn.Module) -> "Sequential":
        r"""Appends a given module to the end.

        Args:
            module (nn.Module): module to append
        """
        self.add_module(str(len(self)), module)
        return self


class BroadcastReduce(Sequential):
    """Broadcast an input to parallel modules and reduce.
    This module is a shorthand for::

        co.Sequential(co.Broadcast(), co.Parallel(*args), co.Reduce(reduce))

    For instance, it can be used to succinctly create a continual 3D Inception Module::

        def norm_relu(module, channels):
            return co.Sequential(
                module,
                nn.BatchNorm3d(channels),
                nn.ReLU(),
            )

        inception_module = co.BroadcastReduce(
            co.Conv3d(192, 64, kernel_size=1),
            co.Sequential(
                norm_relu(co.Conv3d(192, 96, kernel_size=1), 96),
                norm_relu(co.Conv3d(96, 128, kernel_size=3, padding=1), 128),
            ),
            co.Sequential(
                norm_relu(co.Conv3d(192, 16, kernel_size=1), 16),
                norm_relu(co.Conv3d(16, 32, kernel_size=5, padding=2), 32),
            ),
            co.Sequential(
                co.MaxPool3d(kernel_size=(1, 3, 3), padding=(0, 1, 1), stride=1),
                norm_relu(co.Conv3d(192, 32, kernel_size=1), 32),
            ),
            reduce="concat",
        )

    Args:
        arg (OrderedDict[str, CoModule]): An OrderedDict or modules to be applied in parallel.
        *args (CoModule): Modules to be applied in parallel.
        reduce (ReductionFuncOrEnum, optional):
            Function used to reduce the parallel outputs.
            Sum or concatenation can be specified by passing "sum" or "concat" respectively.
            Custom reduce functions can also be passed.
            Defaults to "sum".
        auto_delay (bool, optional):
            Automatically add delay to modules in order to match the longest delay.
            Defaults to True.
    """

    @overload
    def __init__(
        self,
        *args: CoModule,
        reduce: ReductionFuncOrEnum = "sum",
        auto_delay=True,
    ) -> None:
        ...  # pragma: no cover

    @overload
    def __init__(
        self,
        arg: "OrderedDict[str, CoModule]",
        reduce: ReductionFuncOrEnum = "sum",
        auto_delay=True,
    ) -> None:
        ...  # pragma: no cover

    def __init__(
        self,
        *args,
        reduce: ReductionFuncOrEnum = "sum",
        auto_delay=True,
    ):
        nn.Module.__init__(self)

        if len(args) == 1 and isinstance(args[0], OrderedDict):
            modules = [(key, module) for key, module in args[0].items()]
        else:
            modules = [(str(idx), module) for idx, module in enumerate(args)]

        assert (
            len(modules) > 1
        ), "You should pass at least two modules for the map-reduce operation to make sense."

        if auto_delay:
            # If there is a delay mismatch, automatically add delay to match the longest
            max_delay = max([m.delay for _, m in modules])
            modules = [
                (
                    key,
                    (
                        Sequential(module, Delay(max_delay - module.delay))
                        if module.delay < max_delay
                        else module
                    ),
                )
                for key, module in modules
            ]

        assert (
            len(set(num_from(getattr(m, "stride", 1)) for _, m in modules)) == 1
        ), f"Expected all modules to have the same stride, but got strides {[(num_from(getattr(m, 'stride', 1))) for _, m in modules]}"

        for key, module in modules:
            self.add_module(key, module)

        self.reduce = nonempty(
            reduce
            if callable(reduce)
            else {
                Reduction.SUM: reduce_sum,
                Reduction.CONCAT: reduce_concat,
                Reduction.MUL: reduce_mul,
                Reduction.MAX: reduce_max,
            }[Reduction(reduce)]
        )

        delays = set(m.delay for m in self)
        self._delay = max(delays)

        receptive_fields = set(m.receptive_field for m in self)
        self._receptive_field = max(receptive_fields)

    def add_module(self, name: str, module: Optional["nn.Module"]) -> None:
        co_add_module(self, name, module)

    @property
    def _state_shape(self):
        return [m._state_shape for m in self]

    @property
    def _dynamic_state_inds(self):
        return [m._dynamic_state_inds for m in self]

    def _forward_step(self, input: torch.Tensor, prev_state: List[State] = None):
        prev_state = prev_state or [None for _ in range(len(self))]
        next_state = prev_state.copy()
        outs = []
        for i, module in enumerate(self):
            out, n_state = module._forward_step(input, prev_state=prev_state[i])
            next_state[i] = n_state
            outs.append(out)
        if all(isinstance(o, Tensor) for o in outs):
            return self.reduce(outs), next_state
        return None, next_state

    def forward_step(self, input: Tensor, update_state=True) -> Tensor:
        outs = []
        for m in self:
            with temporary_parameter(m, "call_mode", "forward_step"):
                outs.append(m(input, update_state=update_state))  # == m.forward_step
        if all(isinstance(o, Tensor) for o in outs):
            return self.reduce(outs)

        return None

    def forward_steps(self, input: Tensor, pad_end=False, update_state=True) -> Tensor:
        outs = []
        for m in self:
            with temporary_parameter(m, "call_mode", "forward_steps"):
                # m.forward_steps
                outs.append(m(input, pad_end=pad_end, update_state=update_state))

        return self.reduce(outs)

    def forward(self, input: Tensor) -> Tensor:
        outs = []
        for m in self:
            with temporary_parameter(m, "call_mode", "forward"):
                outs.append(apply_forward(m, input))

        return self.reduce(outs)

    @property
    def receptive_field(self) -> int:
        return self._receptive_field

    @property
    def delay(self) -> int:
        return self._delay

    @property
    def stride(self) -> Tuple[int]:
        return getattr(next(iter(self)), "stride", (1,))

    @property
    def padding(self) -> Tuple[int]:
        return (max(getattr(m, "padding", (0,))[0] for m in self),)

    def clean_state(self):
        for m in self:
            if hasattr(m, "clean_state"):
                m.clean_state()

    def extra_repr(self):
        return f"reduce={self.reduce.__name__}"


def Residual(
    module: CoModule,
    temporal_fill: PaddingMode = None,
    reduce: Reduction = "sum",
    residual_shrink: Union[bool, str] = False,
) -> BroadcastReduce:
    """Residual connection wrapper for input.

    This module produces a short form of BroadCast reduce with one delay stream::

        conv = co.Conv3d(32, 32, kernel_size=3, padding=1)
        res1 = co.BroadcastReduce(conv, co.Delay(2), reduce="sum")
        res2 = co.Residual(conv)

        x = torch.randn(1, 32, 5, 5, 5)
        assert torch.equal(res1(x), res2(x))

    Args:
        module (CoModule): module to which a residual should be added.
        temporal_fill (PaddingMode, optional): temporal fill type in delay. Defaults to None.
        reduce (Reduction, optional): Reduction function. Defaults to "sum".
        residual_shrink (bool, optional):
            Set residual to shrink its forward to match the temporal dimension reduction of the wrapped module.
            Options: "centered", "lagging" or True: Centered residual shrink;
                     "lagging": lagging shrink. Defaults to False.
                     "leading": leading shrink, i.e. no delay during forward_step(s).

    Returns:
        BroadcastReduce: BroadcastReduce module with residual.
    """
    assert num_from(getattr(module, "stride", 1)) == 1, (
        "The simple `Residual` only works for modules with temporal stride=1. "
        "Complex residuals can be achieved using `BroadcastReduce` or the `Broadcast`, `Parallel`, and `Reduce` modules."
    )
    temporal_fill = temporal_fill or getattr(
        module, "temporal_fill", PaddingMode.REPLICATE.value
    )
    delay = module.delay
    equal_padding = module.receptive_field - num_from(module.padding) * 2 == 1
    if equal_padding:
        residual_shrink = False

    if residual_shrink in {True, "centered"}:
        assert delay % 2 == 0, "Auto-shrink only works for even-number delays."
        delay = delay // 2

    if residual_shrink == "leading":
        res = Skip(delay)
    else:
        res = Delay(delay, temporal_fill, auto_shrink=residual_shrink)

    return BroadcastReduce(
        res,  # Residual first yields easier broadcasting in reduce functions
        module,
        reduce=reduce,
        auto_delay=False,
    )


class Conditional(FlattenableStateDict, CoModule, nn.Module):
    """Module wrapper for conditional invocations at runtime.

    For instance, it can be used to apply a softmax if the module isn't training::

        net = co.Sequential()

        def not_training(module, x):
            return not net.training

        net.append(co.Conditional(not_training, torch.nn.Softmax(dim=1)))

    Args:
        predicate (Callable[[CoModule, Tensor], bool]):
            Function used to evaluate whether on module or the other should be invoked.
        on_true: CoModule: Module to invoke on True.
        on_false: Optional[CoModule]: Module to invoke on False. If no module is passed, execution is skipped.

    """

    def __init__(
        self,
        predicate: Callable[[CoModule, Tensor], bool],
        on_true: CoModule,
        on_false: CoModule = None,
    ):
        from continual.convert import continual  # Break cyclical import

        assert callable(predicate), "The pased function should be callable."
        if not isinstance(on_true, CoModule):
            on_true = continual(on_true)
        if not (isinstance(on_false, CoModule) or on_false is None):
            on_false = continual(on_false)

        nn.Module.__init__(self)

        self.predicate = predicate

        # Ensure modules have the same delay
        self._delay = max(on_true.delay, getattr(on_false, "delay", 0))
        self._receptive_field = max(
            on_true.receptive_field, getattr(on_false, "receptive_field", 1)
        )

        self.add_module(
            "0",
            on_true
            if on_true.delay == self._delay
            else Sequential(Delay(self._delay - on_true.delay), on_true),
        )

        if on_false is not None:
            self.add_module(
                "1",
                on_false
                if on_false.delay == self._delay
                else Sequential(Delay(self._delay - on_false.delay), on_false),
            )

    def forward(self, input: Tensor) -> Tensor:
        if self.predicate(self, input):
            return apply_forward(self._modules["0"], input)
        elif "1" in self._modules:
            return apply_forward(self._modules["1"], input)
        return input

    def forward_step(self, input: Tensor, update_state=True) -> Tensor:
        if self.predicate(self, input):
            return self._modules["0"].forward_step(input, update_state)
        elif "1" in self._modules:
            return self._modules["1"].forward_step(input, update_state)
        return input

    def _forward_step(
        self, input: Tensor, prev_state: Optional[State] = None
    ) -> Tuple[Tensor, Optional[State]]:
        prev_state = prev_state or [None, None]
        if self.predicate(self, input):
            return self._modules["0"]._forward_step(input, prev_state[0])
        elif "1" in self._modules:
            return self._modules["1"]._forward_step(input, prev_state[1])
        return input, prev_state

    def forward_steps(self, input: Tensor, pad_end=False, update_state=True) -> Tensor:
        if self.predicate(self, input):
            return self._modules["0"].forward_steps(input)
        elif "1" in self._modules:
            return self._modules["1"].forward_steps(input)
        return input

    @property
    def _state_shape(self):
        return [m._state_shape for m in self._modules.values()]

    @property
    def _dynamic_state_inds(self):
        return [m._dynamic_state_inds for m in self._modules.values()]

    @property
    def delay(self) -> int:
        return self._delay

    @property
    def receptive_field(self) -> int:
        return self._receptive_field

    def extra_repr(self):
        return f"predicate={function_repr(self.predicate)}"



""" PyTorch Lightning compatible logging """

import logging
from functools import wraps
from typing import Callable

__all__ = ["getLogger"]


def _process_rank():
    try:  # pragma: no cover
        import pytorch_lightning as pl

        try:
            import horovod.torch as hvd

            hvd.init()
            return hvd.rank()

        except ModuleNotFoundError:
            return pl.utilities.rank_zero_only.rank

    except ModuleNotFoundError:
        return 0


process_rank = _process_rank()


def once(fn: Callable):
    mem = set()

    @wraps(fn)
    def wrapped(*args, **kwargs):
        h = hash((args, str(kwargs)))
        if h in mem:
            return
        mem.add(h)
        return fn(*args, **kwargs)

    return wrapped


def if_rank_zero(fn):
    @wraps(fn)
    def wrapped(*args, **kwargs):
        global process_rank
        if process_rank == 0:
            fn(*args, **kwargs)

    return wrapped


def getLogger(name, log_once=False):
    logger = logging.getLogger(name)
    if log_once:
        logger._log = once(logger._log)
    logger._log = if_rank_zero(logger._log)
    return logger



import functools
import itertools
from abc import ABC
from enum import Enum
from typing import List, Optional, Tuple, Union

import torch
import torch.utils.hooks as hooks
from torch import Tensor
from torch.nn.modules.module import (
    _global_backward_hooks,
    _global_forward_hooks,
    _global_forward_pre_hooks,
)

__all__ = ["CoModule", "call_mode"]

# First element is a value buffer, while others are indexes
State = Tuple[Tensor, Tensor, Tensor]


class PaddingMode(Enum):
    REPLICATE = "replicate"
    ZEROS = "zeros"
    NEG_INF = "neg_inf"


# Not compatible with torch.jit:
# class CallMode(Enum):
#     FORWARD = "forward"
#     FORWARD_STEPS = "forward_steps"
#     FORWARD_STEP = "forward_step"

CALL_MODES = {
    "forward": torch.tensor(0),
    "forward_steps": torch.tensor(1),
    "forward_step": torch.tensor(2),
}


def _callmode(cm) -> torch.Tensor:
    """_summary_

    Args:
        cm (Union[str, int, torch.Tensor]): Identifier for call mode

    Returns:
        torch.Tensor: validated call_mode
    """
    if isinstance(cm, str):
        cm = CALL_MODES[cm.lower()]
    elif isinstance(cm, int):
        cm = torch.tensor(int)
    return cm


class _CallModeContext(object):
    """Context-manager which temporarily specifies a call_mode

    When the call_mode context is used, the ``__call__`` function of continual modules with be set accordingly

    Example::

        sequence = torch.randn(1, 3, 10)
        step = sequence[:, :, -1]

        forward_output = module(sequence)  # Calls `forward`

        with co.call_mode("forward_step"):
            forward_step_output = module(step)  # Calls `forward_step`

        forward_output = module(sequence)  # Calls `forward`
    """

    def __init__(self):
        self.cur = _callmode("forward")
        self.prev = None

    def __call__(self, value: Union[str, int, torch.Tensor]):
        self.prev = self.cur
        self.cur = _callmode(value)
        return self

    def __enter__(self):
        pass

    def __exit__(self, *args, **kwargs):
        self.cur = self.prev
        self.prev = None


call_mode = _CallModeContext()


def _clone_first(state: State) -> State:
    # return (state[0].clone(), state[1], state[2])
    return (state[0].clone(),) + state[1:]
    # return (state[0].clone(),) + state[1:]


class CoModule(ABC):
    """Base class for continual modules.
    Deriving from this class provides base-functionality and enforces
    the implementation of necessary methods.

    Attributes:
        receptive_field (int): Temporal receptive field of the module.
        delay (int): Number of step inputs to observe before the modules produces valid outputs.
        stride (Tuple[int,...]): (Spatio)-temporal stride.
        padding (Tuple[int,...]): (Spatio)-temporal padding.

    """

    receptive_field: int = 1
    stride: Tuple[int, ...] = (1,)
    padding: Tuple[int, ...] = (0,)
    _make_padding = torch.zeros_like
    _state_shape: int = 0
    _dynamic_state_inds: List[bool] = []
    _call_mode = _callmode("forward")

    def __init_subclass__(cls) -> None:
        CoModule._validate_class(cls)

    @staticmethod
    def _validate_class(cls):
        for fn, description in [
            ("forward_step", "forward computation for a single temporal step"),
            (
                "forward_steps",
                "forward computation for multiple temporal step",
            ),
            (
                "forward",
                "a forward computation which is identical to a regular non-continual forward.",
            ),
            ("get_state", "a retrieval of the internal state."),
            ("set_state", "an update of the internal state."),
            ("clean_state", "an internal state clean-up."),
        ]:
            assert callable(
                getattr(cls, fn, None)
            ), f"{cls} should implement a `{fn}` function which performs {description} to satisfy the CoModule interface."

        for prop in {"delay", "receptive_field"}:
            assert type(getattr(cls, prop, None)) in {
                int,
                torch.Tensor,
                property,
            }, f"{cls} should implement a `{prop}` property to satisfy the CoModule interface."

        for prop in {"stride", "padding", "_state_shape", "_dynamic_state_inds"}:
            assert type(getattr(cls, prop, None)) in {
                int,
                property,
                tuple,
                list,
            }, f"{cls} should implement a `{prop}` property to satisfy the CoModule interface."

    @staticmethod
    def is_valid(module):
        try:
            CoModule._validate_class(module)
        except AssertionError:
            return False
        return True

    def get_state(self) -> Optional[State]:
        """Get model state.

        Returns:
            Optional[State]: A State tuple if the model has been initialised and otherwise None.
        """
        ...  # pragma: no cover

    def set_state(self, state: State):
        """Set model state

        Args:
            state (State): State tuple to set as new internal internal state
        """
        ...  # pragma: no cover

    def clean_state(self):
        """Clean model state, resetting the network memory."""
        ...  # pragma: no cover

    @property
    def delay(self) -> int:
        return self.receptive_field - 1 - self.padding[0]

    def forward_step(
        self, input: Tensor, update_state: bool = True
    ) -> Optional[Tensor]:
        """Performs a forward computation for a single frame and (optionally) updates internal states accordingly.
        This function performs efficient continual inference.

        Illustration::

            O+S O+S O+S O+S   (O: output, S: updated internal state)
             ↑   ↑   ↑   ↑
             N   N   N   N    (N: network module)
             ↑   ↑   ↑   ↑
             I   I   I   I    (I: input frame)

        Args:
            input (Tensor): Layer input.
            update_state (bool): Whether internal state should be updated during this operation.

        Returns:
            Optional[Tensor]: Step output. This will be a placeholder while the module initializes and every (stride - 1) / stride.
        """
        state = self.get_state()
        if not update_state and state:
            state = _clone_first(state)
        output, state = self._forward_step(input, state)
        if update_state and state:
            self.set_state(state)
        return output

    def forward_steps(
        self, input: Tensor, pad_end: bool = False, update_state=True
    ) -> Optional[Tensor]:
        """Performs a forward computation across multiple time-steps while updating internal states for continual inference (if update_state=True).
        Start-padding is always accounted for, but end-padding is omitted per default in expectance of the next input step. It can be added by specifying pad_end=True. If so, the output-input mapping the exact same as that of forward.

        Illustration::

                    O            (O: output)
                    ↑
            -----------------    (-: aggregation)
            O  O+S O+S O+S  O    (O: output, S: updated internal state)
            ↑   ↑   ↑   ↑   ↑
            N   N   N   N   N    (N: network module)
            ↑   ↑   ↑   ↑   ↑
            P   I   I   I   P    (I: input frame, P: padding)

        Args:
            input (Tensor): Layer input.
            pad_end (bool): Whether results for temporal padding at sequence end should be included.
            update_state (bool): Whether internal state should be updated during this operation.

        Returns:
            Optional[Tensor]: Layer output
        """
        return self._forward_steps_impl(input, pad_end, update_state)

    def _forward_steps_impl(
        self, input: Tensor, pad_end: bool = False, update_state: bool = True
    ) -> Optional[Tensor]:
        """Forward computation for multiple steps with state initialisation

        Args:
            module (CoModule): Continual module.
            input (Tensor): Layer input.
            pad_end (bool): Whether results for temporal padding at sequence end should be included.
            update_state (bool): Whether internal state should be updated during this operation.

        Returns:
            Tensor: Layer output
        """
        outs = []
        state: Optional[State] = self.get_state()

        if not update_state and state is not None:
            state = _clone_first(state)

        for t in range(input.shape[2]):
            o, state = self._forward_step(input[:, :, t], state)
            if isinstance(o, Tensor):
                outs.append(o)

        if update_state and state is not None:
            self.set_state(state)

        if pad_end:
            # Don't save state for the end-padding
            opt_state = self.get_state()
            if opt_state is not None:
                state = _clone_first(opt_state)

            for t, i in enumerate(
                [self._make_padding(input[:, :, -1]) for _ in range(self.padding[0])]
            ):
                o, state = self._forward_step(i, state)
                if isinstance(o, Tensor):
                    outs.append(o)

        if len(outs) == 0:
            return None  # pragma: no cover

        return torch.stack(outs, dim=2)

    def forward(self, input: Tensor) -> Tensor:
        """Performs a forward computation over multiple time-steps.
        This function is identical to the corresponding module in _torch.nn_,
        ensuring cross-compatibility. Moreover, it's handy for efficient training
        on clip-based data.

        Illustration::

                    O            (O: output)
                    ↑
                    N            (N: network module)
                    ↑
            -----------------    (-: aggregation)
            P   I   I   I   P    (I: input frame, P: padding)

        Args:
            input (Tensor): Network input.
        """
        ...  # pragma: no cover

    def warm_up(self, step_shape: List[int]):
        """Warms up the model state with a dummy input.
        The initial `self.delay` steps will produce results, but they will be inexact.

        To warm up the model with a user-defined data, pass the data to forward_steps::

            net.forward_steps(user_data)

        Args:
            step_shape (Sequence[int]): input shape with which to warm the model up, including batch size.
        """
        step_shape = (*step_shape[:2], self.delay, *step_shape[2:])
        dummy = self._make_padding(torch.zeros(step_shape, dtype=torch.float))
        self.forward_steps(dummy)

    @property
    def call_mode(self) -> torch.Tensor:
        return self._call_mode

    @call_mode.setter
    def call_mode(self, value):
        self._call_mode = _callmode(value)
        if hasattr(self, "__len__"):
            for m in self:
                if hasattr(m, "call_mode"):
                    m.call_mode = self._call_mode

    def _call_impl(self, *input, **kwargs):  # noqa: C901  # pragma: no cover
        """Modified version torch.nn.Module._call_impl

        Returns:
            [type]: [description]
        """
        _call_mode = call_mode.cur if call_mode.prev is not None else self.call_mode
        forward_call = {
            (True, _callmode("forward")): self._slow_forward,
            (False, _callmode("forward")): self.forward,
            (False, _callmode("forward_steps")): self.forward_steps,
            (False, _callmode("forward_step")): self.forward_step,
        }[(bool(torch._C._get_tracing_state()), _call_mode)]

        # If we don't have any hooks, we want to skip the rest of the logic in
        # this function, and just call forward.
        if not (
            self._backward_hooks
            or self._forward_hooks
            or self._forward_pre_hooks
            or _global_backward_hooks
            or _global_forward_hooks
            or _global_forward_pre_hooks
        ):
            return forward_call(*input, **kwargs)
        # Do not call functions when jit is used
        full_backward_hooks, non_full_backward_hooks = [], []
        if self._backward_hooks or _global_backward_hooks:
            full_backward_hooks, non_full_backward_hooks = self._get_backward_hooks()
        if _global_forward_pre_hooks or self._forward_pre_hooks:
            for hook in itertools.chain(
                _global_forward_pre_hooks.values(), self._forward_pre_hooks.values()
            ):
                result = hook(self, input)
                if result is not None:
                    if not isinstance(result, tuple):
                        result = (result,)
                    input = result

        bw_hook = None
        if full_backward_hooks:
            bw_hook = hooks.BackwardHook(self, full_backward_hooks)
            input = bw_hook.setup_input_hook(input)

        result = forward_call(*input, **kwargs)
        if _global_forward_hooks or self._forward_hooks:
            for hook in itertools.chain(
                _global_forward_hooks.values(), self._forward_hooks.values()
            ):
                hook_result = hook(self, input, result)
                if hook_result is not None:
                    result = hook_result

        if bw_hook:
            result = bw_hook.setup_output_hook(result)

        # Handle the non-full backward hooks
        if non_full_backward_hooks:
            var = result
            while not isinstance(var, torch.Tensor):
                if isinstance(var, dict):
                    var = next((v for v in var.values() if isinstance(v, torch.Tensor)))
                else:
                    var = var[0]
            grad_fn = var.grad_fn
            if grad_fn is not None:
                for hook in non_full_backward_hooks:
                    wrapper = functools.partial(hook, self)
                    functools.update_wrapper(wrapper, hook)
                    grad_fn.register_hook(wrapper)
                self._maybe_warn_non_full_backward_hook(input, result, grad_fn)

        return result

    __call__ = _call_impl

    @staticmethod
    def build_from(
        module: torch.nn.Module,
        *args,
        **kwargs,
    ) -> "CoModule":
        """Copy parameters and weights from a non-continual module and
        build the corresponding continual version.

        Args:
            module (torch.nn.Module): Module from which to copy variables and weights

        Returns:
            CoModule: Continual Module with the parameters and weights of the passed module.
        """



import io
from typing import Any, Dict, List, Mapping, Optional, Sequence, Tuple, Union

import torch
import torch._C._onnx as _C_onnx

from continual.module import CoModule
from continual.utils import flatten

__all__ = ["export", "OnnxWrapper"]


def export(
    model: Union[torch.nn.Module, torch.jit.ScriptModule, torch.jit.ScriptFunction],
    args: Union[Tuple[Any, ...], torch.Tensor],
    f: Union[str, io.BytesIO],
    export_params: bool = True,
    verbose: bool = False,
    training: _C_onnx.TrainingMode = _C_onnx.TrainingMode.EVAL,
    input_names: Optional[Sequence[str]] = None,
    output_names: Optional[Sequence[str]] = None,
    operator_export_type: _C_onnx.OperatorExportTypes = _C_onnx.OperatorExportTypes.ONNX,
    opset_version: Optional[int] = None,
    do_constant_folding: bool = True,
    # dynamic_axes: Optional[  # Handled automatically
    #     Union[Mapping[str, Mapping[int, str]], Mapping[str, Sequence[int]]]
    # ] = None,
    keep_initializers_as_inputs: Optional[bool] = None,
    custom_opsets: Optional[Mapping[str, int]] = None,
    # export_modules_as_functions: Union[bool, Collection[Type[torch.nn.Module]]] = False,
):
    r"""Exports a model into ONNX format.
    If ``model`` is not a :class:`torch.jit.ScriptModule` nor a
    :class:`torch.jit.ScriptFunction`, this runs
    ``model`` once in order to convert it to a TorchScript graph to be exported
    (the equivalent of :func:`torch.jit.trace`). Thus this has the same limited support
    for dynamic control flow as :func:`torch.jit.trace`.
    Args:
        model (:class:`torch.nn.Module`, :class:`torch.jit.ScriptModule` or :class:`torch.jit.ScriptFunction`):
            the model to be exported.
        args (tuple or torch.Tensor):
            args can be structured either as:
            1. ONLY A TUPLE OF ARGUMENTS::
                args = (x, y, z)
            The tuple should contain model inputs such that ``model(*args)`` is a valid
            invocation of the model. Any non-Tensor arguments will be hard-coded into the
            exported model; any Tensor arguments will become inputs of the exported model,
            in the order they occur in the tuple.
            2. A TENSOR::
                args = torch.Tensor([1])
            This is equivalent to a 1-ary tuple of that Tensor.
            3. A TUPLE OF ARGUMENTS ENDING WITH A DICTIONARY OF NAMED ARGUMENTS::
                args = (
                    x,
                    {
                        "y": input_y,
                        "z": input_z
                    }
                )
            All but the last element of the tuple will be passed as non-keyword arguments,
            and named arguments will be set from the last element. If a named argument is
            not present in the dictionary, it is assigned the default value, or None if a
            default value is not provided.
            .. note::
                If a dictionary is the last element of the args tuple, it will be
                interpreted as containing named arguments. In order to pass a dict as the
                last non-keyword arg, provide an empty dict as the last element of the args
                tuple. For example, instead of::
                    torch.onnx.export(
                        model,
                        (
                            x,
                            # WRONG: will be interpreted as named arguments
                            {y: z}
                        ),
                        "test.onnx.pb"
                    )
                Write::
                    torch.onnx.export(
                        model,
                        (
                            x,
                            {y: z},
                            {}
                        ),
                        "test.onnx.pb"
                    )
        f: a file-like object (such that ``f.fileno()`` returns a file descriptor)
            or a string containing a file name.  A binary protocol buffer will be written
            to this file.
        export_params (bool, default True): if True, all parameters will
            be exported. Set this to False if you want to export an untrained model.
            In this case, the exported model will first take all of its parameters
            as arguments, with the ordering as specified by ``model.state_dict().values()``
        verbose (bool, default False): if True, prints a description of the
            model being exported to stdout. In addition, the final ONNX graph will include the
            field ``doc_string``` from the exported model which mentions the source code locations
            for ``model``. If True, ONNX exporter logging will be turned on.
        training (enum, default TrainingMode.EVAL):
            * ``TrainingMode.EVAL``: export the model in inference mode.
            * ``TrainingMode.PRESERVE``: export the model in inference mode if model.training is
                False and in training mode if model.training is True.
            * ``TrainingMode.TRAINING``: export the model in training mode. Disables optimizations
                which might interfere with training.
        input_names (list of str, default empty list): names to assign to the
            input nodes of the graph, in order.
        output_names (list of str, default empty list): names to assign to the
            output nodes of the graph, in order.
        operator_export_type (enum, default OperatorExportTypes.ONNX):
            * ``OperatorExportTypes.ONNX``: Export all ops as regular ONNX ops
                (in the default opset domain).
            * ``OperatorExportTypes.ONNX_FALLTHROUGH``: Try to convert all ops
                to standard ONNX ops in the default opset domain. If unable to do so
                (e.g. because support has not been added to convert a particular torch op to ONNX),
                fall back to exporting the op into a custom opset domain without conversion. Applies
                to `custom ops <https://pytorch.org/tutorials/advanced/torch_script_custom_ops.html>`_
                as well as ATen ops. For the exported model to be usable, the runtime must support
                these non-standard ops.
            * ``OperatorExportTypes.ONNX_ATEN``: All ATen ops (in the TorchScript namespace "aten")
                are exported as ATen ops (in opset domain "org.pytorch.aten").
                `ATen <https://pytorch.org/cppdocs/#aten>`_ is PyTorch's built-in tensor library, so
                this instructs the runtime to use PyTorch's implementation of these ops.
                .. warning::
                    Models exported this way are probably runnable only by Caffe2.
                    This may be useful if the numeric differences in implementations of operators are
                    causing large differences in behavior between PyTorch and Caffe2 (which is more
                    common on untrained models).
            * ``OperatorExportTypes.ONNX_ATEN_FALLBACK``: Try to export each ATen op
                (in the TorchScript namespace "aten") as a regular ONNX op. If we are unable to do so
                (e.g. because support has not been added to convert a particular torch op to ONNX),
                fall back to exporting an ATen op. See documentation on OperatorExportTypes.ONNX_ATEN for
                context.
                For example::
                    graph(%0 : Float):
                    %3 : int = prim::Constant[value=0]()
                    # conversion unsupported
                    %4 : Float = aten::triu(%0, %3)
                    # conversion supported
                    %5 : Float = aten::mul(%4, %0)
                    return (%5)
                Assuming ``aten::triu`` is not supported in ONNX, this will be exported as::
                    graph(%0 : Float):
                    %1 : Long() = onnx::Constant[value={0}]()
                    # not converted
                    %2 : Float = aten::ATen[operator="triu"](%0, %1)
                    # converted
                    %3 : Float = onnx::Mul(%2, %0)
                    return (%3)
                If PyTorch was built with Caffe2 (i.e. with ``BUILD_CAFFE2=1``), then
                Caffe2-specific behavior will be enabled, including special support
                for ops are produced by the modules described in
                `Quantization <https://pytorch.org/docs/stable/quantization.html>`_.
                .. warning::
                    Models exported this way are probably runnable only by Caffe2.
        opset_version (int, default 14): The version of the
            `default (ai.onnx) opset <https://github.com/onnx/onnx/blob/master/docs/Operators.md>`_
            to target. Must be >= 7 and <= 16.
        do_constant_folding (bool, default True): Apply the constant-folding optimization.
            Constant-folding will replace some of the ops that have all constant inputs
            with pre-computed constant nodes.
        keep_initializers_as_inputs (bool, default None): If True, all the
            initializers (typically corresponding to parameters) in the
            exported graph will also be added as inputs to the graph. If False,
            then initializers are not added as inputs to the graph, and only
            the non-parameter inputs are added as inputs.
            This may allow for better optimizations (e.g. constant folding) by
            backends/runtimes.
            If ``opset_version < 9``, initializers MUST be part of graph
            inputs and this argument will be ignored and the behavior will be
            equivalent to setting this argument to True.
            If None, then the behavior is chosen automatically as follows:
            * If ``operator_export_type=OperatorExportTypes.ONNX``, the behavior is equivalent
                to setting this argument to False.
            * Else, the behavior is equivalent to setting this argument to True.
        custom_opsets (dict[str, int], default empty dict): A dict with schema:
            * KEY (str): opset domain name
            * VALUE (int): opset version
            If a custom opset is referenced by ``model`` but not mentioned in this dictionary,
            the opset version is set to 1. Only custom opset domain name and version should be
            indicated through this argument.
        export_modules_as_functions (bool or set of type of nn.Module, default False): Flag to enable
            exporting all ``nn.Module`` forward calls as local functions in ONNX. Or a set to indicate the
            particular types of modules to export as local functions in ONNX.
            This feature requires ``opset_version`` >= 15, otherwise the export will fail. This is because
            ``opset_version`` < 15 implies IR version < 8, which means no local function support.
            Module variables will be exported as function attributes. There are two categories of function
            attributes.
            1. Annotated attributes: class variables that have type annotations via
            `PEP 526-style <https://www.python.org/dev/peps/pep-0526/#class-and-instance-variable-annotations>`_
            will be exported as attributes.
            Annotated attributes are not used inside the subgraph of ONNX local function because
            they are not created by PyTorch JIT tracing, but they may be used by consumers
            to determine whether or not to replace the function with a particular fused kernel.
            2. Inferred attributes: variables that are used by operators inside the module. Attribute names
            will have prefix "inferred::". This is to differentiate from predefined attributes retrieved from
            python module annotations. Inferred attributes are used inside the subgraph of ONNX local function.
            * ``False`` (default): export ``nn.Module`` forward calls as fine grained nodes.
            * ``True``: export all ``nn.Module`` forward calls as local function nodes.
            * Set of type of nn.Module: export ``nn.Module`` forward calls as local function nodes,
                only if the type of the ``nn.Module`` is found in the set.
    Raises:
        :class:`torch.onnx.errors.CheckerError`: If the ONNX checker detects an invalid ONNX graph.
        :class:`torch.onnx.errors.UnsupportedOperatorError`: If the ONNX graph cannot be exported because it
            uses an operator that is not supported by the exporter.
        :class:`torch.onnx.errors.OnnxExporterError`: Other errors that can occur during export.
            All errors are subclasses of :class:`errors.OnnxExporterError`.
    """
    assert isinstance(
        model, CoModule
    ), f"The passed model of type {type(model)} should be a CoModule"
    omodel = OnnxWrapper(model)
    torch.onnx.export(
        model=omodel,
        args=args,
        f=f,
        export_params=export_params,
        verbose=verbose,
        training=training,
        input_names=input_names + omodel.state_input_names,
        output_names=output_names + omodel.state_output_names,
        operator_export_type=operator_export_type,
        opset_version=opset_version,
        do_constant_folding=do_constant_folding,
        dynamic_axes={
            **{i: {0: "batch"} for i in input_names},
            **{o: {0: "batch"} for o in output_names},
            **omodel.state_dynamic_axes,
        },
        keep_initializers_as_inputs=keep_initializers_as_inputs,
        custom_opsets=custom_opsets,
    )


def _shape_list(lst, shape, idx=0):
    if isinstance(shape, int):
        return lst[idx : idx + shape], idx + shape

    assert hasattr(shape, "__len__")
    ret = []
    for s in shape:
        o, idx = _shape_list(lst, s, idx)
        ret.append(o)
    return ret, idx


class OnnxWrapper(torch.nn.Module):
    """Collapses input args and flattens output args.
    This is necessary as the ``dynamic_axes`` arg for
    :py:meth:`torch.onnx.export` doesn't accept nested Tuples.
    Args:
        model: A co.CoModule
    """

    def __init__(self, model: CoModule):
        super().__init__()
        self.model = model

    def forward(self, x: torch.Tensor, *states: torch.Tensor):
        shaped_state, _ = _shape_list(states, self.model._state_shape)

        out, next_states = self.model._forward_step(x, shaped_state)
        return (out, *flatten(next_states, remove_none=False))

    @staticmethod
    def _i2o_name(i_name: str) -> str:
        return f"n{i_name}"

    @property
    def state_input_names(self) -> List[str]:
        return [f"s{i}" for i in range(sum(flatten(self.model._state_shape)))]

    @property
    def state_output_names(self) -> List[str]:
        return [self._i2o_name(s) for s in self.state_input_names]

    @property
    def state_dynamic_axes(self) -> Dict[str, List[int]]:
        isdyn = flatten(self.model._dynamic_state_inds)
        ins = {sn: {0: "batch"} for sn, i in zip(self.state_input_names, isdyn) if i}
        outs = {self._i2o_name(k): v for k, v in ins.items()}
        return {**ins, **outs}



from collections import OrderedDict
from enum import Enum
from functools import partial, reduce
from typing import Callable, Optional, Sequence, Tuple, Union

import torch
from torch import Tensor, nn

from continual.delay import State as DelayState

from .closure import Identity, Lambda
from .container import BroadcastReduce, Residual, Sequential
from .delay import Delay, PaddingMode
from .linear import Linear
from .multihead_attention import (
    RetroactiveMultiheadAttention,
    SingleOutputMultiheadAttention,
)

__all__ = [
    "TransformerEncoder",
    "TransformerEncoderLayerFactory",
    "SingleOutputTransformerEncoderLayer",
    "RetroactiveTransformerEncoderLayer",
]


class MhaType(Enum):
    """Type of Multi-head Attention
    Supported tupes are:
    - RETROACTIVE:      RetroactiveMultiheadAttention
    - SINGLE_OUTPUT:    SingleOutputMultiheadAttention
    - REGULAR:          nn.MultiheadAttention
    """

    RETROACTIVE = "retroactive"
    SINGLE_OUTPUT = "single_output"
    SINGLE_OUTPUT_FORWARD = "single_output_forward"
    REGULAR = "regular"


class SelectOrDelay(Delay):
    """Select a temporal index during forward, or delay correspondingly during forward_step(s)"""

    def forward(self, x: Tensor) -> Optional[Tensor]:
        assert len(x.shape) >= 3  # N, C, T
        return x[:, :, -1 - self._delay].unsqueeze(2)


class RetroactiveUnity(Delay):
    """Unity mapping during forward. During forward_step(s), a single-to-many mapping is assumed,
    and all cached values are output."""

    def __init__(
        self,
        delay: int,
        temporal_fill: PaddingMode = "zeros",
        auto_shrink: bool = False,
        time_dim=-1,
    ):
        """Initialise Delay block

        Args:
            delay (int): the number of steps to delay an output.
            temporal_fill (PaddingMode, optional): Temporal state initialisation mode ("zeros" or "replicate"). Defaults to "zeros".
            auto_shrink (int, optional): Whether to shrink the temporal dimension of the feature map during forward.
                This is handy for residuals that are parallel to modules which reduce the number of temporal steps. Defaults to False.
            time_dim (int, optional): Which dimension to concatenate step outputs along
        """
        self.time_dim = time_dim
        Delay.__init__(self, delay, temporal_fill, auto_shrink)

    def init_state(
        self,
        first_output: Tensor,
    ) -> DelayState:
        padding = self._make_padding(first_output)
        state_buffer = torch.stack([padding for _ in range(self.delay + 1)], dim=0)
        state_index = torch.tensor(-self.delay)
        return state_buffer, state_index

    def _forward_step(
        self, input: Tensor, prev_state: DelayState
    ) -> Tuple[Tensor, DelayState]:
        if prev_state is None:
            buffer, index = self.init_state(input)
        else:
            buffer, index = prev_state

        # Update state
        buffer[index % (self.delay + 1)] = input
        new_index = index + 1
        if new_index > 0:
            new_index = new_index % self.delay

        # Get output
        output = None
        if index >= 0:
            output = buffer.clone().roll(shifts=int(-index - 1), dims=0)
            idx = (
                self.time_dim + len(output.shape)
                if self.time_dim < 0
                else self.time_dim
            )
            output = output.permute(
                list(range(1, idx + 1)) + [0] + list(range(idx + 1, len(output.shape)))
            )

        return output, (buffer, new_index)


class RetroactiveLambda(Lambda):
    """
    Lambda wrapper for functions that are applied after retroactive modules.
    """

    def forward(self, input: Tensor) -> Tensor:
        return Lambda.forward(self, input)

    def forward_step(self, input: Tensor, *args, **kwargs) -> Tensor:
        return self.forward(input)

    def _forward_step(self, input: Tensor, prev_state=None, *args, **kwargs) -> Tensor:
        return self.forward(input), prev_state

    def forward_steps(self, input: Tensor, *args, **kwargs) -> Tensor:
        return torch.stack(
            [self.forward(input[:, :, t]) for t in range(input.shape[2])], dim=2
        )

    @staticmethod
    def build_from(
        fn: Callable[[Tensor], Tensor], takes_time=False
    ) -> "RetroactiveLambda":
        return RetroactiveLambda(fn, takes_time)


class NaiveResidual(nn.Module):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def forward(self, x):
        return self.fn(x) + x


def sum_last_pairs(inputs: Sequence[Tensor]) -> Tensor:
    if inputs[0].shape != inputs[1].shape:
        T_min = min(inputs[i].shape[2] for i in range(len(inputs)))
        inputs = [inp[:, :, -T_min:] for inp in inputs]
    return reduce(torch.Tensor.add, inputs[1:], inputs[0])


def SingleOutputTransformerEncoderLayer(
    d_model: int,
    nhead: int,
    dim_feedforward: int = 2048,
    dropout: float = 0.1,
    activation: Union[nn.Module, Callable[[Tensor], Tensor]] = nn.functional.relu,
    layer_norm_eps: float = 1e-5,
    # batch_first: bool = True,
    # norm_first: bool = False,
    device=None,
    dtype=None,
    sequence_len: int = None,
    single_output_forward=False,
    query_index: int = -1,
):
    """Continual Single-output Transformer Encoder layer.

    Contrary to the ``torch.nn.TransformerEncoderLayer``, this layer only computes the attention
    for the last query during ``forward_step``. The behavior during ``forward`` is controllable
    with the ``single_output_forward`` parameter.

    The continual formulation of the Transformer Encoder Layer was proposed by Hedegaard et al.
    in "Continual Transformers: Redundancy-Free Attention for Online Inference".
    https://arxiv.org/abs/2201.06268 (paper) https://www.youtube.com/watch?v=gy802Tlp-eQ (video).

    .. note::
        In order to handle positional encoding correctly for continual input streams, the
        :class:`RecyclingPositionalEncoding` should be used together with this module.

    Args:
        d_model: the number of expected features in the input (required).
        nhead: the number of heads in the multiheadattention models (required).
        dim_feedforward: the dimension of the feedforward network model (default=2048).
        dropout: the dropout value (default=0.1).
        activation: the activation function of the intermediate layer, can be a string
            ("relu" or "gelu") or a unary callable. Default: relu.
        layer_norm_eps: the eps value in layer normalization components (default=1e-5).
        device: torch device to initialize layer on. Defaults to None.
        dtype: datatype of layer parameters. Defaults to None.
        sequence_len: length of token-sequence to perform attention across. Defaults to None.
        single_output_forward: whether to restrict the attention to the last token during forward. Defaults to False.
        query_index: the sequence position index to compute the attention for.

    Examples::

        encoder_layer = co.SingleOutputTransformerEncoderLayer(
            d_model=512, nhead=8, sequence_len=32, dropout=0.0
        )
        x = torch.rand(10, 512, 32)  # (N, E, T)

        # corresponds to torch.nn.TransformerEncoderLayer
        out = encoder_layer.forward(x)

        # continual inference API
        firsts = encoder_layer.forward_steps(x[:,:,:-1])
        last = encoder_layer.forward_step(x[:,:,-1])

        assert firsts is None  # The module first needs to observe ``sequence_len`` values
        assert torch.allclose(out[:,:,-1], last, atol=1e-6)

    """
    assert (
        sequence_len > 0
    ), "Please provide a positive integer value as sequence length."

    factory_kwargs = {"device": device, "dtype": dtype}
    norm1 = nn.LayerNorm(d_model, eps=layer_norm_eps, **factory_kwargs)
    norm2 = nn.LayerNorm(d_model, eps=layer_norm_eps, **factory_kwargs)

    mha = SingleOutputMultiheadAttention(
        embed_dim=d_model,
        num_heads=nhead,
        dropout=dropout,
        bias=True,
        batch_first=True,
        embed_dim_second=True,
        query_index=query_index,
        device=device,
        dtype=dtype,
        sequence_len=sequence_len,
        forward_returns_attn_mask=False,
        single_output_forward=single_output_forward,
    )

    ff = Sequential(
        OrderedDict(
            [
                (
                    "linear1",
                    Linear(d_model, dim_feedforward, channel_dim=1, **factory_kwargs),
                ),
                (
                    "activation",
                    activation,
                ),
                (
                    "dropout",
                    nn.Dropout(dropout),
                ),
                (
                    "linear2",
                    Linear(dim_feedforward, d_model, channel_dim=1, **factory_kwargs),
                ),
                (
                    "dropout2",
                    nn.Dropout(dropout),
                ),
            ]
        )
    )

    return Sequential(
        BroadcastReduce(
            OrderedDict(
                [
                    (
                        "residual",
                        SelectOrDelay(mha.delay)
                        if single_output_forward
                        else Identity(),
                    ),
                    (
                        "self_attn",
                        mha,
                    ),
                ]
            ),
            reduce=sum_last_pairs,
            auto_delay=False,
        ),
        Sequential(
            OrderedDict(
                [
                    ("norm1", Lambda(norm1, takes_time=False)),
                    ("_ff_block", Residual(ff)),
                    ("norm2", Lambda(norm2, takes_time=False)),
                ]
            )
        ),
    )


def RetroactiveTransformerEncoderLayer(
    d_model: int,
    nhead: int,
    dim_feedforward: int = 2048,
    dropout: float = 0.1,
    activation: Union[nn.Module, Callable[[Tensor], Tensor]] = nn.functional.relu,
    layer_norm_eps: float = 1e-5,
    # batch_first: bool = True,
    # norm_first: bool = False,
    device=None,
    dtype=None,
    sequence_len: int = None,
):
    """Continual Retroactive Transformer Encoder layer.

    When a new token is received, it computes the updated attention values corresponding
    to prior tokens as well.

    The continual formulation of the Transformer Encoder Layer was proposed by Hedegaard et al.
    in "Continual Transformers: Redundancy-Free Attention for Online Inference".
    https://arxiv.org/abs/2201.06268 (paper) https://www.youtube.com/watch?v=gy802Tlp-eQ (video).

    .. note::
        In order to handle positional encoding correctly for continual input streams, the
        :class:`RecyclingPositionalEncoding` should be used together with this module.

    Args:
        d_model: the number of expected features in the input (required).
        nhead: the number of heads in the multiheadattention models (required).
        dim_feedforward: the dimension of the feedforward network model (default=2048).
        dropout: the dropout value (default=0.1).
        activation: the activation function of the intermediate layer, can be a string
            ("relu" or "gelu") or a unary callable. Default: relu.
        layer_norm_eps: the eps value in layer normalization components (default=1e-5).
        device: torch device to initialize layer on. Defaults to None.
        dtype: datatype of layer parameters. Defaults to None.
        sequence_len: length of token-sequence to perform attention across. Defaults to None.

    Examples::

        encoder_layer = co.RetroactiveTransformerEncoderLayer(
            d_model=512, nhead=8, sequence_len=32, dropout=0.0
        )
        x = torch.rand(10, 512, 32)  # (N, E, T)

        # corresponds to torch.nn.TransformerEncoderLayer
        out = encoder_layer.forward(x)

        # continual inference API
        firsts = encoder_layer.forward_steps(x[:,:,:-1])
        last = encoder_layer.forward_step(x[:,:,-1])

        assert firsts is None  # The module first needs to observe ``sequence_len`` values
        assert torch.allclose(out, last, atol=1e-6)

    """
    assert (
        sequence_len > 0
    ), "Please provide a positive integer value as sequence length."
    factory_kwargs = {"device": device, "dtype": dtype}
    norm1 = nn.LayerNorm(d_model, eps=layer_norm_eps, **factory_kwargs)
    norm2 = nn.LayerNorm(d_model, eps=layer_norm_eps, **factory_kwargs)

    mha = RetroactiveMultiheadAttention(
        embed_dim=d_model,
        num_heads=nhead,
        dropout=dropout,
        bias=True,
        batch_first=True,
        embed_dim_second=True,
        device=device,
        dtype=dtype,
        sequence_len=sequence_len,
        forward_returns_attn_mask=False,
    )

    ff = Sequential(
        OrderedDict(
            [
                ("linear1", Linear(d_model, dim_feedforward, **factory_kwargs)),
                ("activation", Lambda(activation, takes_time=True)),
                ("dropout", nn.Dropout(dropout)),
                ("linear2", Linear(dim_feedforward, d_model, **factory_kwargs)),
                ("dropout2", nn.Dropout(dropout)),
            ]
        )
    )

    return Sequential(
        BroadcastReduce(
            OrderedDict(
                [
                    ("residual", RetroactiveUnity(mha.delay)),
                    ("self_attn", mha),
                ]
            ),
            reduce="sum",
            auto_delay=False,
        ),
        RetroactiveLambda(
            nn.Sequential(
                OrderedDict(
                    [
                        ("norm1", norm1),
                        ("_ff_block", NaiveResidual(ff)),
                        ("norm2", norm2),
                    ]
                )
            )
        ),
    )


# TODO: impl
def StepLocalTransformerEncoderLayer(
    d_model: int,
    nhead: int,
    dim_feedforward: int = 2048,
    dropout: float = 0.1,
    activation: Union[nn.Module, Callable[[Tensor], Tensor]] = nn.functional.relu,
    layer_norm_eps: float = 1e-5,
    # batch_first: bool = True,
    # norm_first: bool = False,
    device=None,
    dtype=None,
    sequence_len: int = None,
):
    ...


def TransformerEncoderLayerFactory(
    d_model: int,
    nhead: int,
    dim_feedforward: int = 2048,
    dropout: float = 0.1,
    activation: Union[nn.Module, Callable[[Tensor], Tensor], str] = nn.functional.relu,
    layer_norm_eps: float = 1e-5,
    # batch_first: bool = True,
    # norm_first: bool = False,
    device=None,
    dtype=None,
    sequence_len: int = None,
) -> Callable[[MhaType], Sequential]:
    """Defines the hyper-parameters of Continual Transformer Encoder layers, where each layer
    contains feed forward networks and continual multi-head attentions as proposed by
    Vaswani et al. in "Attention is all you need".

    It can produce either a :class:`SingleOutputTransformerEncoderLayer` or a
    :class:`RetroactiveTransformerEncoderLayer`.
    These were proposed by Hedegaard et al. in
    "Continual Transformers: Redundancy-Free Attention for Online Inference".
    https://arxiv.org/abs/2201.06268 (paper) https://www.youtube.com/watch?v=gy802Tlp-eQ (video).

    Args:
        d_model: the number of expected features in the input (required).
        nhead: the number of heads in the multiheadattention models (required).
        dim_feedforward: the dimension of the feedforward network model (default=2048).
        dropout: the dropout value (default=0.1).
        activation: the activation function of the intermediate layer, can be a string
            ("relu" or "gelu") or a unary callable. Default: relu.
        layer_norm_eps: the eps value in layer normalization components (default=1e-5).
        device: torch device to initialize layer on. Defaults to None.
        dtype: datatype of layer parameters. Defaults to None.
        sequence_len: length of token-sequence to perform attention across. Defaults to None.

    Returns:
        Callable[[Union[str,MhaType]], Sequential]: Factory function return the layer module given
            the desired MHA type (one of "retroactive", "single_output", "single_output_forward", and "regular").

    Examples::

        encoder_layer = co.TransformerEncoderLayerFactory(d_model=512, nhead=8, sequence_len=32)
        transformer_encoder = co.TransformerEncoder(encoder_layer, num_layers=2)
        src = torch.rand(10, 512, 32)
        out = transformer_encoder(src)
    """
    if activation in {"relu", "gelu"}:
        activation = {
            "relu": nn.functional.relu,
            "gelu": nn.functional.relu,
        }[activation]

    def TransformerEncoderLayer(mha_type: MhaType):
        factory_fn = {
            MhaType.RETROACTIVE: RetroactiveTransformerEncoderLayer,
            MhaType.SINGLE_OUTPUT: SingleOutputTransformerEncoderLayer,
            MhaType.SINGLE_OUTPUT_FORWARD: partial(
                SingleOutputTransformerEncoderLayer, single_output_forward=True
            ),
            MhaType.REGULAR: StepLocalTransformerEncoderLayer,
        }[MhaType(mha_type)]

        return factory_fn(
            d_model,
            nhead,
            dim_feedforward,
            dropout,
            activation,
            layer_norm_eps,
            # batch_first
            # norm_first
            device,
            dtype,
            sequence_len,
        )

    return TransformerEncoderLayer


class TransformerEncoder(Sequential):
    """Continual Transformer Encoder is a stack of N encoder layers.

    The continual formulation of the Transformer Encoder was proposed by Hedegaard et al.
    in "Continual Transformers: Redundancy-Free Attention for Online Inference".
    https://arxiv.org/abs/2201.06268 (paper) https://www.youtube.com/watch?v=gy802Tlp-eQ (video).

    .. note::
        This class deviates from the Pytorch implementation in the following ways:
        1) `encoder_layer` parameter takes a factory functor, TransformerEncoderLayerFactory
        2) `mask` and `src_key_padding_mask` are not supported currently.

    .. note::
        The efficiency gains of ``forward_step`` compared to ``forward`` is highly dependent
        on the chosen ``num_layers``. Here, a lower ``num_layers`` is most efficient.
        Accordingly, we recommend increasing ``d_model``, ``nhead``, and ``dim_feedforward``
        of the :class:`TransformerEncoderLayerFactory` rather than increasing ``num_layers`` if larger
        models are desired. Keeping the parameter-count equal, this was found to work well
        for regular Transformer Encoders as well (https://arxiv.org/pdf/2210.00640.pdf).

    .. note::
        In order to handle positional encoding correctly for continual input streams, the
        :class:`RecyclingPositionalEncoding` should be used together with this module.

    Args:
        encoder_layer: An instance of :class:`TransformerEncoderLayerFactory`.
        num_layers: the number of sub-encoder-layers in the encoder (required).
        norm: the layer normalization component (optional).

    Examples::

        encoder_layer = co.TransformerEncoderLayerFactory(d_model=512, nhead=8, sequence_len=32)
        transformer_encoder = co.TransformerEncoder(encoder_layer, num_layers=2)
        src = torch.rand(10, 512, 32)
        out = transformer_encoder(src)
    """

    def __init__(
        self,
        encoder_layer: Callable[[MhaType, Optional[bool]], Sequential],
        num_layers: int,
        norm: nn.Module = None,
    ):
        layers = []
        if num_layers == 1:
            layers.append(encoder_layer(MhaType.SINGLE_OUTPUT))
        else:
            layers.append(encoder_layer(MhaType.RETROACTIVE))
            for _ in range(1, num_layers - 1):
                layers.append(
                    RetroactiveLambda(encoder_layer(MhaType.REGULAR), takes_time=True)
                )
            layers.append(
                RetroactiveLambda(
                    encoder_layer(MhaType.SINGLE_OUTPUT_FORWARD), takes_time=True
                )
            )

            def unity(x):
                return x

            def squeeze_last(x):
                return x.squeeze(-1)

            layers.append(
                Lambda(unity, None, squeeze_last, squeeze_last, takes_time=True)
            )

        Sequential.__init__(self, OrderedDict([("layers", Sequential(*layers))]))
        if norm is not None:
            self.add_module("norm", Lambda(norm, takes_time=False))

    @staticmethod
    def build_from(
        trans_enc: nn.TransformerEncoder, sequence_len: int
    ) -> "TransformerEncoder":
        assert isinstance(trans_enc, nn.TransformerEncoder)

        # Create model
        tel = trans_enc.layers[0]
        layer_factory = TransformerEncoderLayerFactory(
            d_model=tel.self_attn.embed_dim,
            nhead=tel.self_attn.num_heads,
            dim_feedforward=tel.linear1.out_features,
            dropout=tel.dropout.p,
            activation=tel.activation,
            layer_norm_eps=tel.norm1.eps,
            device=tel.linear1.weight.device,
            dtype=tel.linear1.weight.dtype,
            sequence_len=sequence_len,
        )
        net = TransformerEncoder(
            layer_factory, num_layers=trans_enc.num_layers, norm=trans_enc.norm
        )

        # Transfer weights
        new_sd = {}

        net_keys = list(net.state_dict().keys())
        match_keys, key_inds = zip(
            *sorted(
                [
                    (
                        k.replace("fn.", "")
                        .replace("_ff_block.", "")
                        .replace(".0.self_attn", ".self_attn"),
                        i,
                    )
                    for i, k in enumerate(net_keys)
                ],
                key=lambda x: x[0],
            )
        )

        reg_keys, weights = zip(
            *sorted(
                [item for item in trans_enc.state_dict().items()], key=lambda x: x[0]
            )
        )

        assert all(
            [
                ".".join(k1.split(".")[-2:]) == ".".join(k2.split(".")[-2:])
                for k1, k2 in zip(match_keys, reg_keys)
            ]
        )

        new_sd = {net_keys[key_inds[i]]: weights[i] for i in range(len(net_keys))}

        net.load_state_dict(new_sd)

        return net



from collections import OrderedDict, abc
from contextlib import contextmanager
from functools import partial, reduce
from inspect import getsource
from numbers import Number
from typing import Any, List, Sequence, Tuple, Union

import torch
from torch import Tensor, nn

from .logging import getLogger

__all__ = [
    "rsetattr",
    "rgetattr",
    "temporary_parameter",
    "flat_state_dict",
    "state_dict",
    "load_state_dict",
    "flatten",
    "num_from",
    "function_repr",
]

logger = getLogger(__name__)


def rsetattr(obj, attr, val):
    pre, _, post = attr.rpartition(".")
    return setattr(rgetattr(obj, pre) if pre else obj, post, val)


def rgetattr(obj, attr, *args):
    def _getattr(obj, attr):
        return getattr(obj, attr, *args)

    return reduce(_getattr, [obj] + attr.split("."))


@contextmanager
def temporary_parameter(obj, attr: str, val):
    do_del = False
    try:
        prev_val = rgetattr(obj, attr)
    except AttributeError:
        do_del = True
        assert (
            attr.count(".") == 0
        ), "Nonexisting attributes can only be set one level deep."

    rsetattr(obj, attr, val)

    yield obj

    if do_del:
        delattr(obj, attr)
    else:
        rsetattr(obj, attr, prev_val)


class _FlatStateDict(object):
    """Context-manager state holder."""

    def __init__(self):
        self.flatten = False

    def __enter__(self):
        self.flatten = True

    def __exit__(self, *args, **kwargs):
        self.flatten = False


flat_state_dict = _FlatStateDict()
"""Context-manager that flattens the state dict of containers.

If a container module was not explicitely named by means of an OrderedDict,
it will attempt to flatten the keys during both the `state_dict` and `load_state_dict` operations.

Example:
    >>> with co.flat_state_dict:
    >>>     sd = module.state_dict()  # All unnamed nested keys are flattened, e.g. "0.weight" -> "weight"
    >>>     module.load_state_dict(sd)  # Automatically unflattened during loading "weight" -> "0.weight"
"""


def state_dict(
    module: nn.Module, destination=None, prefix="", keep_vars=False, flatten=False
) -> "OrderedDict[str, Tensor]":
    """Returns a dictionary containing a whole state of the module.

    Both parameters and persistent buffers (e.g. running averages) are
    included. Keys are corresponding parameter and buffer names.

    Args:
        destination (OrderedDict, optional): a dict to which the state is saved.
        prefix (str, optional): A string prefix for state keys.
        keep_vars (bool, optional): Whether parameters should not be detached.
        flatten (bool, optional): whether the state dict keys are flattened (no numeric keys).

    Returns:
        dict:
            a dictionary containing a whole state of the module

    Example::

        >>> module.state_dict().keys()
        ['bias', 'weight']

    """
    d = nn.Module.state_dict(module, destination, prefix, keep_vars)

    if flatten or flat_state_dict.flatten:
        flat_keys = [
            ".".join(part for part in name.split(".") if not part.isdigit())
            for name in list(d.keys())
        ]
        if len(set(flat_keys)) == len(d.keys()):
            d = OrderedDict(list(zip(flat_keys, d.values())))
        else:  # pragma: no cover
            logger.warning("Unable to flatten state dict consistently.")

    return d


def load_state_dict(
    module: nn.Module,
    state_dict: "OrderedDict[str, Tensor]",
    strict: bool = True,
    flatten=False,
):
    """Copies parameters and buffers from :attr:`state_dict` into
    this module and its descendants. If :attr:`strict` is ``True``, then
    the keys of :attr:`state_dict` must exactly match the keys returned
    by this module's :meth:`~torch.nn.Module.state_dict` function.

    Args:
        state_dict (dict): a dict containing parameters and
            persistent buffers.
        strict (bool, optional): whether to strictly enforce that the keys
            in :attr:`state_dict` match the keys returned by this module's
            :meth:`~torch.nn.Module.state_dict` function. Default: ``True``
        flatten (bool, optional): whether the loaded state dict is flattened (no numeric keys) during loading.


    Returns:
        ``NamedTuple`` with ``missing_keys`` and ``unexpected_keys`` fields:
            * **missing_keys** is a list of str containing the missing keys
            * **unexpected_keys** is a list of str containing the unexpected keys
    """

    if flatten or flat_state_dict.flatten:
        long_keys = nn.Module.state_dict(module, keep_vars=True).keys()

        short2long = {
            ".".join(part for part in key.split(".") if not part.isdigit()): key
            for key in list(long_keys)
        }
        if len(short2long) == len(state_dict):
            state_dict = OrderedDict(
                [
                    (short2long[key], val)
                    for key, val in state_dict.items()
                    if strict or key in short2long
                ]
            )
        else:  # pragma: no cover
            logger.warning(
                "Unable to deduce a consistent flattened state_dict. Loading as regular state_dict."
            )
    #
    return nn.Module.load_state_dict(module, state_dict, strict)


def num_from(
    tuple_or_num: Union[
        torch.Tensor,
        int,
        Tuple[int],
        Tuple[int, int],
        Tuple[int, int, int],
        Tuple[int, int, int, int],
    ],
    dim: int = 0,
):
    """Extract number from dimension or leave as is

    Args:
        tuple_or_num (Union[Number, Tuple[Number, ...]]): _description_
        dim (int, optional): _description_. Defaults to 0.

    Returns:
        Number: Number found along dimension
    """
    if isinstance(tuple_or_num, Number):
        return tuple_or_num

    return tuple_or_num[dim]


def function_repr(fn):
    if fn is None:
        return ""

    if isinstance(fn, nn.Module):
        fn.__name__ = fn.__repr__()

    if isinstance(fn, partial):
        fn = fn.func
    s = fn.__name__
    if s == "<lambda>":
        s = getsource(fn)
        s = s[s.find("lambda") :]

        # first ':'
        i = s.find(":", 0)

        stack = 0
        for c in s[i + 1 :]:
            i += 1
            if c == "(":
                stack += 1
            elif c == ")":
                stack -= 1
            if (stack == 0 and c == ",") or (stack == -1 and c == ")"):
                break

        s = s[:i]
        s = s.rstrip()
    return s


def flatten(lst: Union[Sequence[Any], Any], remove_none=True) -> List[Any]:
    lst = _flatten(lst)
    if remove_none:
        lst = [i for i in lst if i is not None]
    return lst


def _flatten(lst: Union[Sequence[Any], Any]) -> List[Any]:
    if isinstance(lst, abc.Sequence):
        if len(lst) == 0:
            return []
        first, rest = lst[0], lst[1:]
        return _flatten(first) + _flatten(rest)
    else:
        return [lst]


def neginf_like(x: Tensor):
    return -torch.ones_like(x) * float("inf")



from functools import partial
from typing import Callable, Union

import torch
from torch import Tensor, nn

from .module import CoModule
from .utils import function_repr

__all__ = [
    "Lambda",
    "Multiply",
    "Add",
    "Identity",
    "Constant",
    "Zero",
    "One",
]


class Lambda(CoModule, nn.Module):
    """Module wrapper for stateless functions.

    .. note::
        Operations performed in a Lambda are not counted in `ptflops`

    Args:
        fn: Function to be called during forward.
        forward_only_fn: Function to be called only during ``forward``. ``fn`` is used for the other call modes.
        forward_step_only_fn: Function to be called only during ``forward_step``. ``fn`` is used for the other call modes.
        forward_steps_only_fn: Function to be called only during ``forward_steps``. ``fn`` is used for the other call modes.
        forward_only_fn: Function to be called only during forward. ``fn`` is used for the other call modes.
        takes_time: If True, ``fn`` receives all steps, if False, it received one step and no time dimension. Defaults to False.

    Examples::

        x = torch.arange(90).reshape(1,3,30) * 1.0

        # Using named function
        def same_stats_different_values(x):
            return torch.randn_like(x) * x.std() + x.mean()

        same_stats_layer = co.Lambda(same_stats_different_values)
        same_stats_layer(x)

        # Using unnamed function
        mean_layer = co.Lambda(lambda x: torch.ones_like(x) * x.mean())
        mean_layer(x)

        # Using functor
        sigmoid = co.Lambda(torch.nn.Sigmoid())
        sigmoid(x)

    """

    _state_shape = 0
    _dynamic_state_inds = []

    def __init__(
        self,
        fn: Callable[[Tensor], Tensor] = None,
        forward_only_fn=None,
        forward_step_only_fn=None,
        forward_steps_only_fn=None,
        takes_time=False,
    ):
        nn.Module.__init__(self)
        assert callable(fn) or all(
            [
                callable(forward_only_fn),
                callable(forward_step_only_fn),
                callable(forward_steps_only_fn),
            ]
        ), "Either fn or all of forward_only_fn, forward_step_only_fn, and forward_steps_only_fn should be callable."

        self.fn = fn
        self.forward_only_fn = forward_only_fn
        self.forward_step_only_fn = forward_step_only_fn
        self.forward_steps_only_fn = forward_steps_only_fn
        self.takes_time = takes_time

    @staticmethod
    def build_from(
        fn: Callable[[Tensor], Tensor],
        forward_only_fn=None,
        forward_step_only_fn=None,
        forward_steps_only_fn=None,
        takes_time=False,
    ) -> "Lambda":
        return Lambda(
            fn, forward_only_fn, forward_step_only_fn, forward_steps_only_fn, takes_time
        )

    def __repr__(self) -> str:
        s = self.__class__.__name__ + "("
        if callable(self.fn):
            s += f"{function_repr(self.fn)}"
        if callable(self.forward_only_fn):
            if callable(self.fn):
                s += ", "
            s += f"{function_repr(self.forward_only_fn)}"
        if callable(self.forward_step_only_fn):
            s += f", {function_repr(self.forward_step_only_fn)}"
        if callable(self.forward_steps_only_fn):
            s += f", {function_repr(self.forward_steps_only_fn)}"
        if self.takes_time:
            s += ", takes_time=True"
        s += ")"
        return s

    def forward(self, input: Tensor) -> Tensor:
        if self.forward_only_fn is not None:
            return self.forward_only_fn(input)

        if self.takes_time:
            return self.fn(input)

        return torch.stack(
            [self.fn(input[:, :, t]) for t in range(input.shape[2])], dim=2
        )

    def forward_steps(self, input: Tensor, pad_end=False, update_state=True) -> Tensor:
        if self.forward_steps_only_fn is not None:
            return self.forward_steps_only_fn(input)

        if self.takes_time:
            return self.fn(input)

        return torch.stack(
            [self.fn(input[:, :, t]) for t in range(input.shape[2])], dim=2
        )

    def forward_step(self, input: Tensor, update_state=True) -> Tensor:
        return self._forward_step(input)[0]

    def _forward_step(self, input: Tensor, prev_state=None) -> Tensor:
        if self.forward_step_only_fn is not None:
            return self.forward_step_only_fn(input), prev_state

        if self.takes_time:
            input = input.unsqueeze(dim=2)
        output = self.fn(input)
        if self.takes_time:
            output = output.squeeze(dim=2)

        return output, prev_state


def _multiply(x: Tensor, factor: Union[float, int, Tensor]):
    return x * factor


def Multiply(factor: float) -> Lambda:
    r"""Applies an scaling transformation to the incoming data: :math:`y = ax`.

    Args:
        factor (float): Number to multiply with.
    """
    fn = partial(_multiply, factor=factor)
    return Lambda(fn, takes_time=True)


def _add(x: Tensor, constant: Union[float, int, Tensor]):
    return x + constant


def Add(constant: float) -> Lambda:
    r"""Applies an additive translation to the incoming data: :math:`y = x + a`.

    Args:
        constant (float): Number to add.
    """
    """Create Lambda with addition function"""
    fn = partial(_add, constant=constant)
    return Lambda(fn, takes_time=True)


def _unity(x: Tensor):
    return x


def Identity(*args, **kwargs) -> Lambda:
    """A placeholder identity operator that is argument-insensitive.

    Args:
        args: any argument (unused)
        kwargs: any keyword argument (unused)

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        m = co.Identity(54, unused_argument1=0.1, unused_argument2=False)
        input = torch.randn(128, 20)
        output = m(input)
        assert output.size() == torch.Size([128, 20])

    """
    return Lambda(_unity, takes_time=True)


def Constant(constant: float) -> Lambda:
    """Returns ``constant * torch.ones_like(input)``.

    Arguments:
        constant: Constant value to return.
    """
    return Lambda(lambda x: constant * torch.ones_like(x), takes_time=True)


def Zero() -> Lambda:
    """Returns ``torch.zeros_like(input)``."""
    return Lambda(torch.zeros_like, takes_time=True)


def One() -> Lambda:
    """Returns ``torch.ones_like(input)``."""
    return Lambda(torch.ones_like, takes_time=True)



from typing import Tuple

import torch
from torch import Tensor

from .module import CoModule

State = Tuple[Tensor]

__all__ = ["Skip"]


class Skip(CoModule, torch.nn.Module):
    """Skip a number of input steps.

    Arguments:
        steps: The number of steps to skip.
    """

    @property
    def _state_shape(self):
        return 1

    @property
    def _dynamic_state_inds(self):
        return [False]

    def __init__(self, steps: int):
        assert steps > 0
        self.steps = steps

        super(Skip, self).__init__()
        self.register_buffer("state_index", torch.tensor(-self.steps), persistent=False)

    def clean_state(self):
        self.state_index = torch.tensor(-self.steps)

    def get_state(self):
        return (self.state_index,)

    def set_state(self, state: State):
        self.state_index = state[0]

    def _forward_step(self, input: Tensor, prev_state: State) -> Tuple[Tensor, State]:
        if prev_state is None:
            index = torch.tensor(-self.steps)
        else:
            index = prev_state[0]

        if index >= 0:
            output = input
            new_index = index
        else:
            output = None
            new_index = index + 1

        return output, (new_index,)

    def forward(self, input: Tensor) -> Tensor:
        return input[:, :, self.steps :]

    @property
    def receptive_field(self) -> int:
        return 1

    @property
    def delay(self) -> int:
        return 0

    @property
    def stride(self) -> Tuple[int]:
        return (1,)

    def extra_repr(self):
        return f"{self.steps}"
